/*
   See pages 220-222, Figure 7.6, from "Principles of Parallel Programming".
*/
#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"
#include <windows.h>  // used for Sleep() only

#include "2D_SOR.h"
#include "initial_data_2.h"

#define Top    0
#define Left   0
#define Right  (Cols-1)
#define Bottom (Rows-1)

#define  NorthPE(i) ((i)-Cols)
#define  SouthPE(i) ((i)+Cols)
#define  EastPE(i)  ((i)+1)
#define  WestPE(i)  ((i)-1)

double **temp;

#define RootProcess 0
int tag = 1;

/*
  Command line args: matrix size, number of rows of processes,
                                  number of columns of processes, iterations
*/
int main(int argc, char** argv)
{
   int myID;          /* rank of process */
   int p;             /* number of processes */
   int n;             /* number of rows (and columns) of data */
   int Rows;          /* number of rows of processes */
   int Cols;          /* number of columns of processes */
   double globalDelta;

   int Height, Width;
   double *buffer, *blockBuffer;
   MPI_Status status;
   int i, j, row, col;

   MPI_Init(&argc, &argv);
   MPI_Comm_size(MPI_COMM_WORLD, &p);
   MPI_Comm_rank(MPI_COMM_WORLD, &myID);

   /*** read and check command line arguments ***/
   /*** make sure there are four command line arguments ***/
   /*** be sure Rows * Cols = p (the number of processors) ***/
   /*** be sure Rows and Cols each divide n (or data would be lost) ***/
   if (argc == 5)
   {
      n = atoi(argv[1]);           /* number of rows (and columns) of data */
      Rows = atoi(argv[2]);        /* number of rows of processes */
      Cols = atoi(argv[3]);        /* number of columns of processess */
      iterations = atoi(argv[4]);  /* determines when to stop */
   }
   /* check all the argument constraints */
   if ( (argc != 5)      ||
        (Rows*Cols != p) ||
        (n % Rows != 0)  ||
        (n % Cols != 0)  )
   {
      if (myID == RootProcess) /* RootProcess prints usage message */
      {
         printf("usage: %s n Rows Cols threshold\n", argv[0]);
         printf("       Where Rows and Cols both divide n, and\n");
         printf("       Rows * Cols = number of processors.\n");
      }
      MPI_Finalize();
      return 0;   /* halt all the processes */
   }

   printf("rank = %d, processes = %d, n = %d, Rows = %d, Cols = %d, iterations = %d\n",
           myID, p, n, Rows, Cols, iterations);

   Height = (n/Rows)+2;
   Width  = (n/Cols)+2;

   /* Initialize data, distribute it among the various processes */
   if(myID == RootProcess)
   {
      InitializeData(&new, &val, n);
      blockBuffer = (double*)malloc(Height*Width*sizeof(double));

      /* iterate through the 2-dim "array" of blocks */
      for(row = 0; row < Rows; row++)
         for(col = 0; col < Cols; col++)
            if ( !((row == 0) && (col == 0)) ) /* skip RootProcess's block */
            {
               /* move data from val into blockBuffer */
               for(i = 0; i < Height; i++)
                  for(j = 0; j < Width; j++)
                  {
                     blockBuffer[i*Width+j] = val[(row*(Height-2))+i][(col*(Width-2))+j];
                  }

               /* send the other processes their share of the data */
               printf("sending original data to rank=%d\n", (row*Cols)+col );
               MPI_Send(blockBuffer, Height*Width, MPI_DOUBLE, (row*Cols)+col, tag, MPI_COMM_WORLD);
            }
   }
   else /* non root process receives data */
   {
      /* this process needs to allocate its own space for val and new */
      /* see http://c-faq.com/aryptr/dynmuldimary.html */
      val = (double**)malloc(Height*sizeof(double*));
      val[0] = (double*)malloc(Height * Width * sizeof(double));
      for(i = 1; i < Height; i++)
         val[i] = val[0] + (i * Width);

      /* receive my share of the data from the root process */
      MPI_Recv(val[0], Height*Width, MPI_DOUBLE, RootProcess, tag, MPI_COMM_WORLD, &status);
      printf("rank = %d received original data \n", myID );

      /* make the array new a copy of the array val */
      new = (double**)malloc(Height*sizeof(double*));
      new[0] = (double*)malloc(Height * Width * sizeof(double));
      for(i = 1; i < Height; i++)
         new[i] = new[0] + (i * Width);
      for(i = 0; i < Height; i++)
         for(j = 0; j < Width; j++)
            new[i][j] = val[i][j];
   }

   buffer = (double*)malloc((Height-2) * sizeof(double));
   /* determine the row and col of this process in the "array" of processes */
   row = myID / Cols;
   col = myID % Cols;
   //printf("rank = %d, row = %d, col = %d\n", myID, row, col);

   do
   { /*
      * Send data to four neighbors
      */
      if(row != Top)   /* Send North */
      {
         MPI_Send(&val[1][1], Width-2, MPI_DOUBLE, NorthPE(myID), tag, MPI_COMM_WORLD);
      }

      if(col != Right)  /* Send East */
      {
         for(i=1; i < Height-1; i++)
         {
            buffer[i-1] = val[i][Width-2];
         }
         MPI_Send(buffer, Height-2, MPI_DOUBLE, EastPE(myID), tag, MPI_COMM_WORLD);
      }

      if(row != Bottom)   /* Send South */
      {
         MPI_Send(&val[Height-2][1], Width-2, MPI_DOUBLE, SouthPE(myID), tag, MPI_COMM_WORLD);
      }

      if(col != Left)  /* Send West */
      {
         for(i=1; i < Height-1; i++)
         {
            buffer[i-1] = val[i][1];
         }
         MPI_Send(buffer, Height-2, MPI_DOUBLE, WestPE(myID), tag, MPI_COMM_WORLD);
      }

      /*
       * Receive messages
       */
      if(row != Top)   /* Receive from North */
      {
         MPI_Recv(&val[0][1], Width-2, MPI_DOUBLE, NorthPE(myID), tag, MPI_COMM_WORLD, &status);
      }

      if(col != Right)  /* Receive from East */
      {
         MPI_Recv(buffer, Height-2, MPI_DOUBLE, EastPE(myID), tag, MPI_COMM_WORLD, &status);
         for(i=1; i < Height-1; i++)
         {
            val[i][Width-1] = buffer[i-1];
         }
      }

      if(row != Bottom)   /* Receive from South */
      {
         MPI_Recv(&val[Height-1][1], Width-2, MPI_DOUBLE, SouthPE(myID), tag, MPI_COMM_WORLD, &status);
      }

      if(col != Left)  /* Receive from West */
      {
         MPI_Recv(buffer, Height-2, MPI_DOUBLE, WestPE(myID), tag, MPI_COMM_WORLD, &status);
         for(i=1; i < Height-1; i++)
         {
            val[i][0] = buffer[i-1];
         }
      }

      delta = 0.0;
      /*** update array new using data in array val ***/
      for(i = 1; i < Height-1; i++)
      {
         for(j = 1; j < Width-1; j++)
         {
            new[i][j] = (val[i-1][j] + val[i][j+1] +
                         val[i+1][j] + val[i][j-1]) / 4;
            delta = Max(delta, Abs(new[i][j] - val[i][j]));
         }
      }

      /* Find maximum delta */
      MPI_Allreduce(&delta, &globalDelta, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);
      temp = new;   /* prepare for next iteration */
      new = val;
      val = temp;
/*
      // Use this block of code for debugging purposes
      printf("rank = %d, delta = %f, globalDelta = %f\n", myID, delta, globalDelta);
      // Print out the contents of this process's block.
      // Use parameters p=4, n=14, Rows=2, Cols=2, threshold=0.001
      // to get a good visualization of the computation.
      for(i = 0; i < Height; i++)
      {  for(j = 0; j < Width; j++)
            printf("% #4.4f ", val[i][j]);
         printf("\n");
      }
      Sleep(2000);
*/
      iterates++;
   }while(iterates < iterations);


   /*** Need to get all of the final results in one place. ***/
   if(myID == RootProcess)
   {  /* gather results from all the other processes */
      /* iterate through the 2-dim "array" of blocks */
      for(row = 0; row < Rows; row++)
      {  for(col = 0; col < Cols; col++)
         {  if ( !((row == 0) && (col == 0)) )
            {  /* receive data from each of the other process */
               MPI_Recv(blockBuffer, Height*Width, MPI_DOUBLE, (row*Cols)+col, tag, MPI_COMM_WORLD, &status);
               printf("RootProcess received final data from rank = %d\n", (row*Cols)+col );
               /* move data from blockBuffer into val */
               for(i = 0; i < Height; i++)
                  for(j = 0; j < Width; j++)
                     val[(row*(Height-2))+i][(col*(Width-2))+j] = blockBuffer[i*Width+j];
            }
         }
      }
      /* write the data to a file so that it can be visualized */
      WriteData("data.raw", val, n);
      printf("total iterations = %d\n", iterates);
      printf("final delta = %g\n", globalDelta);
   }
   else /* send my results to the RootProcess */
   {
      printf("sending final data to RootProcess from rank=%d\n", myID );
      MPI_Send(val[0], Height*Width, MPI_DOUBLE, RootProcess, tag, MPI_COMM_WORLD);
   }

   MPI_Finalize();
   return 0;
}/* main() */
